from blok import blok, InitContext, ExecutionContext, Option
from blok.tree import YamlFile, Repo
from dataclasses import dataclass
from typing import Dict, Any, Protocol

from blok import blok, InitContext, service


@service("live.arkitekt.certer", description="Generate HTTPS certificates for services")
class CerterService(Protocol):

    def retrieve_certs_mount(self) -> str: ...

    def retrieve_depends_on(self) -> list[str]: ...



from importlib.metadata import version
from pydantic import BaseModel, ConfigDict, Field, field_validator
import datetime
from typing import List, Optional, Union, Literal, Dict
from enum import Enum
import semver
import uuid
from arkitekt_next.base_models import Requirement
from string import Formatter
import os


class Manifest(BaseModel):
    identifier: str
    version: str
    author: str
    logo: Optional[str] = None
    entrypoint: str
    scopes: List[str]
    model_config = ConfigDict(extra="forbid", validate_assignment=True)

    @field_validator("version", mode="before")
    def version_must_be_semver(cls, v) -> str:
        """Checks that the version is a valid semver version"""
        if isinstance(v, str):
            try:
                semver.VersionInfo.parse(v)
            except ValueError:
                raise ValueError("Version must be a valid semver version")
        return str(v)

    def to_console_string(self):
        return f"📦 {self.identifier} ({self.version}) by {self.author}"

    def to_builder_dict(self):
        return {
            "identifier": self.identifier,
            "version": self.version,
            "logo": self.logo,
            "scopes": self.scopes,
        }




from pydantic import BaseModel
from typing import Dict, Any
from arkitekt_next.bloks.services.name import NameService
from blok import blok, InitContext, Option, ExecutionContext
from blok.tree import YamlFile
from arkitekt_next.bloks.services.secret import SecretService
import secrets
import namegenerator


@blok(NameService, description="Preformed names that will not change")
class PreformedNamesBlok:
    def __init__(self) -> None:
        self.preformed_names = [namegenerator.gen() for _ in range(100)]
        self.used_names = []

    def preflight(self, init: InitContext, preformed_names: list[str]):
        self.preformed_names = list(preformed_names)
        self.used_names = []

    def retrieve_name(self) -> str:
        name = self.preformed_names.pop()
        self.used_names.append(name)
        return name

    def get_options(self):
        config_path = Option(
            subcommand="preformed_names",
            help="A list of preformed names",
            default=self.preformed_names,
            multiple=True,
        )

        return [config_path]



from importlib import import_module
import os
import shutil
from click import ClickException
import rich_click as click
from arkitekt_next.cli.options import (
    with_documents,
    with_graphql_config,
    with_api_path,
    with_boring,
    with_choose_services,
    with_schemas,
    with_seperate_document_dirs,
)
import yaml
from arkitekt_next.cli.utils import build_relative_dir
from arkitekt_next.cli.vars import get_console, get_manifest
from arkitekt_next.service_registry import get_default_service_registry


@click.command()
@with_seperate_document_dirs
@with_boring
@with_choose_services
@with_graphql_config
@with_api_path
@with_schemas
@with_graphql_config
@with_documents
@click.pass_context
def init(ctx, boring, services, config, documents, schemas, path, seperate_doc_dirs):
    """Initialize code generation for the arkitekt_next app

    Code generation for API's is done with the help of GraphQL Code Generation
    that is powered by turms. This command initializes the code generation for
    the app. It creates the necessary folders and files for the code generation
    to work. It also creates a graphql config file that is used by turms to
    generate the code.

    """
    manifest = get_manifest(ctx)
    console = get_console(ctx)
    
    entrypoint = manifest.entrypoint

    with console.status("Loading entrypoint module..."):
        try:
            import_module(entrypoint)
        except ModuleNotFoundError as e:
            console.print(f"Could not find entrypoint module {entrypoint}")
            raise e

    
    
    app_directory = os.getcwd()

    app_api_path = os.path.join(app_directory, path)
    app_documents = os.path.join(app_directory, "documents")

    app_schemas = os.path.join(app_directory, "schemas")

    os.makedirs(app_documents, exist_ok=True)
    os.makedirs(app_schemas, exist_ok=True)
    os.makedirs(app_api_path, exist_ok=True)

    # Initializing the config
    projects = {}

    registry = get_default_service_registry()

    chosen_services = registry.service_builders

    if services:
        chosen_services = {
            key: service
            for key, service in registry.service_builders.items()
            if key in services
        }
    else:
        service = click.prompt(
            "Choose a service to initialize the project for",
            type=click.Choice(list(chosen_services.keys())),
        )

        chosen_services = {service: chosen_services[service]}

    if os.path.exists(config):
        if click.confirm(
            f"GraphQL Config file already exists. Do you want to merge your choices?"
        ):
            file = yaml.load(open(config, "r"), Loader=yaml.FullLoader)
            projects = file.get("projects", {})
            click.echo(
                f"Merging {','.join(chosen_services.keys())} in {','.join(projects.keys())}..."
            )

    has_done_something = False

    for key, service in chosen_services.items():
        try:

            schema, project = service.get_graphql_schema(), service.get_turms_project()

            if not schema or not project:
                get_console(ctx).print(f"[red]No schema or project found for {key} [/]")
                continue

            if key in projects:
                get_console(ctx).print(f"[red]Project {key} already exists [/]")
                if not click.confirm("Do you want to overwrite it?"):
                    continue

            has_done_something = True

            if documents:
                os.makedirs(os.path.join(app_documents, key), exist_ok=True)
                if seperate_doc_dirs:
                    os.makedirs(
                        os.path.join(app_documents, key, "queries"), exist_ok=True
                    )
                    os.makedirs(
                        os.path.join(app_documents, key, "mutations"), exist_ok=True
                    )
                    os.makedirs(
                        os.path.join(app_documents, key, "subscriptions"), exist_ok=True
                    )

            if schemas:
                out_path = os.path.join(app_schemas, key + ".schema.graphql")
                with open(out_path, "w") as f:
                    f.write(schema)

            if schemas:
                project["schema"] = os.path.join(app_schemas, key + ".schema.graphql")
            if documents:
                project["documents"] = (
                    os.path.join(app_documents, key) + "/**/*.graphql"
                )

            project["extensions"]["turms"]["out_dir"] = path
            project["extensions"]["turms"]["generated_name"] = f"{key}.py"
            del project["extensions"]["turms"]["documents"]
            del project["schema_url"]

            projects[key] = project

        except Exception as e:
            raise ClickException(
                f"Failed to initialize project for {key}. Error: {e}"
            ) from e

    if has_done_something:
        graph_config_path = os.path.join(app_directory, config)
        yaml.safe_dump(
            {"projects": projects}, open(graph_config_path, "w"), sort_keys=False
        )
        get_console(ctx).print(f"Config file written to {graph_config_path}")
    else:
        get_console(ctx).print("No projects initialized")
        get_console(ctx).print("Exiting...")



from blok import blok, InitContext, ExecutionContext, Option
from blok.tree import YamlFile, Repo
from dataclasses import dataclass
from typing import Dict, Any, Protocol, Optional

from blok import blok, InitContext, service
from dataclasses import dataclass


@dataclass
class DBCredentials:
    password: str
    username: str
    host: str
    port: int
    db_name: str
    engine: str = "django.db.backends.postgresql"
    dependency: Optional[str] = None


@service("live.arkitekt.postgres")
class DBService(Protocol):
    def register_db(self, db_name: str) -> DBCredentials: ...



from typing import Dict, Any
import secrets

from blok import blok, InitContext, ExecutionContext, Option
from blok.tree import Repo, YamlFile
from arkitekt_next.bloks.base import BaseArkitektService


@blok("live.arkitekt.rekuest", description="Rekuest is the RPC and Functionaliy hub")
class RekuestBlok(BaseArkitektService):
    def __init__(self) -> None:
        self.dev = False
        self.host = "rekuest"
        self.command = "bash run-debug.sh"
        self.repo = "https://github.com/jhnnsrs/rekuest-server-next"
        self.scopes = {
            "rekuest_agent": "Act as an agent",
            "rekuest_call": "Call other apps with rekuest",
        }
        self.mount_repo = False
        self.build_repo = False
        self.buckets = ["media"]
        self.secret_key = secrets.token_hex(16)
        self.image = "jhnnsrs/rekuest_next:nightly"

    def get_builder(self):
        return "arkitekt.rekuest"

    def build(self, context: ExecutionContext):
        context.docker_compose.set_nested("services", self.host, self.service)



from .admin import AdminService
from .lok import LokService
from .db import DBService
from .redis import RedisService
from .s3 import S3Service
from .config import ConfigService
from .mount import MountService
from .secret import SecretService
from .gateway import GatewayService
from .livekit import LivekitService


__all__ = [
    "AdminService",
    "LokService",
    "DBService",
    "RedisService",
    "S3Service",
    "ConfigService",
    "MountService",
    "SecretService",
    "GatewayService",
    "LivekitService",
    "MountService",
    "ConfigService",
    "SecretService",
]



import datetime
import uuid
from arkitekt_next.utils import create_arkitekt_next_folder
import os
from typing import Optional, List, Dict
from arkitekt_next.cli.types import (
    Manifest,
)
import yaml
import json
import rich_click as click


def load_manifest_yaml(path: str) -> Manifest:
    """Loads a manifest from a yaml file

    Uses yaml safe load to load the manifest from a yaml file
    (to avoid unsafe yaml attributes)

    Parameters
    ----------
    path : str
        The path to the yaml file

    Returns
    -------
    Manifest
        The loaded manifest
    """
    with open(path, "r") as file:
        manifest = yaml.safe_load(file)
        return Manifest(**manifest)


def load_manifest() -> Optional[Manifest]:
    """Loads the manifest from the arkitekt_next folder

    Will load the manifest from the current working directories
    arkitekt_next folder. If no folder exists, it will create one, but
    will not create a manifest.

    Returns
    -------
    Optional[Manifest]
        The loaded manifest, or None if no manifest exists
    """
    path = create_arkitekt_next_folder()
    config_file = os.path.join(path, "manifest.yaml")
    if os.path.exists(config_file):
        return load_manifest_yaml(config_file)
    return None


def write_manifest(manifest: Manifest):
    """Writes a manifest to the arkitekt_next folder

    Will write a manifest to the current working directories
    arkitekt_next folder. If no folder exists, it will create one.


    Parameters
    ----------
    manifest : Manifest
        The manifest to write
    """
    path = create_arkitekt_next_folder()
    config_file = os.path.join(path, "manifest.yaml")

    with open(config_file, "w") as file:
        yaml.safe_dump(
            json.loads(manifest.json(exclude_none=True, exclude_unset=True)),
            file,
            sort_keys=True,
        )



from arkitekt_next import get_current_service_registry
import rich_click as click
from importlib import import_module
from arkitekt_next.apps.types import App
from arkitekt_next.cli.commands.run.utils import import_builder
from arkitekt_next.cli.vars import get_console, get_manifest
from arkitekt_next.cli.options import with_builder
import json
import os


async def run_app(app):
    async with app:
        await app.rekuest.run()


@click.command("prod")
@click.pass_context
@click.option(
    "--pretty",
    "-p",
    help="Should we just output json?",
    is_flag=True,
    default=False,
)
@click.option(
    "--machine-readable",
    "-mr",
    help="Should we just output json?",
    is_flag=True,
    default=False,
)
def requirements(
    ctx,
    pretty: bool,
    machine_readable: bool,
):
    """Checks the requirements of the app

    \n
    """

    manifest = get_manifest(ctx)
    console = get_console(ctx)

    entrypoint = manifest.entrypoint
    identifier = manifest.identifier
    entrypoint_file = f"{manifest.entrypoint}.py"
    os.path.realpath(entrypoint_file)


    entrypoint = manifest.entrypoint

    with console.status("Loading entrypoint module..."):
        try:
            import_module(entrypoint)
        except ModuleNotFoundError as e:
            console.print(f"Could not find entrypoint module {entrypoint}")
            raise e


    service_registry = get_current_service_registry()
    
    

    x = [item.model_dump(by_alias=True) for item in service_registry.get_requirements()]
    

    if machine_readable:
        print("--START_REQUIREMENTS--" + json.dumps(x) + "--END_REQUIREMENTS--")

    else:
        if pretty:
            console.print(json.dumps(x, indent=2))
        else:
            print(json.dumps(x))



from importlib import import_module
import inspect as pyinspect
import rich_click as click

from arkitekt_next.cli.ui import construct_leaking_group
from arkitekt_next.cli.vars import get_console, get_manifest
from rich.panel import Panel

import rich_click as click


def inspect_dangerous_variables(module_path):
    """Inspect the module and return a dictionary of all the variables that are
    not upper case and that are not classes, modules, functions or builtins.

    This is used to check if a module is safe to import. Or if it runs code
    that might be dangegours:

    TODO: This is not the ebst way to do this. We should probably use the ast
    module to parse the module and check for dangerous code. This is a quick
    and dirty solution.

    """

    module = import_module(module_path)

    dangerous_variables = {}

    for key, value in pyinspect.getmembers(module):
        if key.startswith("_"):
            continue
        if pyinspect.isclass(value):
            continue
        if pyinspect.ismodule(value):
            continue
        if pyinspect.isfunction(value):
            continue
        if pyinspect.isbuiltin(value):
            continue

        if type(value) in [str, float, int, bool, list, dict, tuple]:
            if key != key.upper():
                dangerous_variables[key] = value
            continue

    return dangerous_variables


def scan_module(module_path):
    """Scan a module for dangerous variables."""
    return inspect_dangerous_variables(module_path)


@click.command()
@click.pass_context
def variables(ctx):
    """Scans your arkitekt_next app for unsafe variables

    When designing an ArkitektNext app, you should not have variables in your
    global scope, as on potential reloads (like for example in development)
    these variables will be redefined and cause memory leaks and other issues.

    You can use this command to scan your app for dangerous variables. If you
    have any, you should consider moving them into a function or class.

    """

    console = get_console(ctx)
    manifest = get_manifest(ctx)
    entrypoint = manifest.entrypoint

    variables = scan_module(entrypoint)

    if not variables:
        console.print(
            Panel(
                "No dangerous variables found. You are good to go!  🎉",
                style="green",
                border_style="green",
                title="ArkitektNext Scan",
            )
        )
        return

    group = construct_leaking_group(variables)

    panel = Panel(
        group, title="ArkitektNext Scan", expand=True, border_style="red", style="red"
    )
    console.print(panel)



import datetime
import uuid
from arkitekt_next.utils import create_arkitekt_next_folder
import os
from typing import Optional, List, Dict
from arkitekt_next.cli.types import (
    Manifest,
)

from .types import (
    Build,
    BuildsConfigFile,
    Flavour,
    DeploymentsConfigFile,
)
from kabinet.api.schema import (
    InspectionInput,
    AppImageInput,
    DockerImageInput,
    ManifestInput,
)

import yaml
import json
import rich_click as click


def get_builds(selected_run: Optional[str] = None) -> Dict[str, Build]:
    """Will load the builds.yaml file and return a dictionary of builds

    Will load the builds.yaml file and return a dictionary of builds
    where the key is the build_id and the value is the build object.


    Returns
    -------
    Dict[str, Build]
        The loaded builds
    """
    path = create_arkitekt_next_folder()
    config_file = os.path.join(path, "builds.yaml")

    builds = {}

    if os.path.exists(config_file):
        with open(config_file, "r") as file:
            config = BuildsConfigFile(**yaml.safe_load(file))

            # We will only return the builds from the selected run
            selected_run = selected_run or config.latest_build_run

            builds = {
                build.build_id: build
                for build in config.builds
                if build.build_run == selected_run
            }
            return builds
    else:
        raise click.ClickException(
            "Could not find any builds. Please run `arkitekt_next port build` first"
        )


def manifest_to_input(manifest: Manifest) -> ManifestInput:

    return ManifestInput(**manifest.model_dump(by_alias=True))


def generate_build(
    build_run: str,
    build_id: str,
    flavour_name: str,
    flavour: Flavour,
    manifest: Manifest,
    inspection: Optional[InspectionInput],
) -> Build:
    """Generates a build from a builder, build_id and manifest

    Will generate a build from a builder, build_id and manifest,
    and write it to the builds.yaml file in the arkitekt_next folder.


    Parameters
    ----------
    builder : str
        The builder that was used to build the build
    build_id : str
        The build_id of the build
    manifest : Manifest
        The manifest of the build

    Returns
    -------
    Build
        The generated build
    """
    path = create_arkitekt_next_folder()

    config_file = os.path.join(path, "builds.yaml")

    build = Build(
        manifest=manifest_to_input(manifest),
        flavour=flavour_name,
        selectors=flavour.selectors,
        build_id=build_id,
        build_run=build_run,
        description=flavour.description,
        inspection=inspection,
    )

    if os.path.exists(config_file):
        with open(config_file, "r") as file:
            config = BuildsConfigFile(**yaml.safe_load(file))
            config.builds.append(build)
            config.latest_build_run = build_run
    else:
        config = BuildsConfigFile(builds=[build], latest_build_run=build_run)

    with open(config_file, "w") as file:
        yaml.safe_dump(
            json.loads(
                config.json(exclude_none=True, exclude_unset=True, by_alias=True)
            ),
            file,
            sort_keys=True,
        )

    return build


def get_deployments() -> DeploymentsConfigFile:
    """Loads the deployments.yaml file and returns the deployments

    Will load the deployments.yaml file and return the deployments
    as a DeploymentsConfigFile object. If no deployments.yaml file
    exists, it will return an empty DeploymentsConfigFile object.

    Returns
    -------
    DeploymentsConfigFile
        The deployments as a DeploymentsConfigFile object
    """
    path = create_arkitekt_next_folder()
    config_file = os.path.join(path, "deployments.yaml")
    if os.path.exists(config_file):
        with open(config_file, "r") as file:
            return DeploymentsConfigFile(**yaml.safe_load(file))
    else:
        return DeploymentsConfigFile()


def generate_deployment(
    deployment_run: str,
    build: Build,
    image: str,
) -> AppImageInput:
    """Generates a deployment from a build and an image

    Parameters
    ----------

    build : Build
        The build that should be deployed
    image: str
        The image that is the actuall deployment of the build
    with_definitions: bool:
        Should we generated and inspect definitions to bundle with
        the deployment?

    Returns:
    ------
    Deployment: The created deployment

    """

    path = create_arkitekt_next_folder()

    config_file = os.path.join(path, "deployments.yaml")

    app_image = AppImageInput(
        appImageId=uuid.uuid4().hex,
        manifest=build.manifest,
        flavourName=build.flavour,
        selectors=build.selectors,
        inspection=build.inspection,
        image=DockerImageInput(imageString=image, buildAt=datetime.datetime.now()),
    )

    if os.path.exists(config_file):
        with open(config_file, "r") as file:
            config = DeploymentsConfigFile(**yaml.safe_load(file))
            config.app_images.append(app_image)
            config.latest_app_image = app_image.app_image_id
    else:
        config = DeploymentsConfigFile(
            app_images=[app_image], latest_app_image=app_image.app_image_id
        )

    with open(config_file, "w") as file:
        yaml.safe_dump(
            json.loads(config.model_dump_json(exclude_none=True, by_alias=True)),
            file,
            sort_keys=True,
        )

    return app_image



from arkitekt_next.apps.types import App, Fakts, Herre, Manifest
from typing import List, Callable, Dict, Any


class QtApp(App):
    """An app that is built with the easy builder"""

    hooks: Dict[str, List[Callable]] = {
        "on_start": [],
        "on_stop": [],
        "on_error": [],
        "on_message": [],
        "on_warning": [],
        "on_info": [],
        "on_debug": [],
        "on_enter": [],
    }

    def register_hook(self, hook_name: str, hook: Callable):
        """Register a hook"""
        self.hooks[hook_name].append(hook)

    def register(self, *args, **kwargs):
        """Register a function with the app (not in the Qt event loop)

        This is useful for functions that do not need access to the Qt event loop
        and can be run in a separate thread. This is the default behavior for
        functions registered with the app.


        """
        self.services["rekuest"].register(*args, **kwargs)

    def register_in_qt_loop(self, function: Callable, **kwargs):
        """Register a function in the Qt event loop

        This is useful for functions that need to be resolved in the Qt event loop and are
        able to block the main thread (e.g. when prompting a necessary user input, or
        when displaying a blocking dialog).

        """

        from rekuest_next.qt.builders import qtinloopactifier

        return self.services["rekuest"].register(
            function, actifier=qtinloopactifier, **kwargs
        )

    def register_with_qt_future(self, function: Callable, **kwargs):
        """Register a function with a future that can be resolved in the Qt event loop

        This is useful for functions that need to be resolved in the Qt event loop, but
        might cause a blocking call if they are run in the same thread.

        Example:

        ```python

            class MyWidget(QtWidgets.QWidget):

                def __init__(self, app: QtApp):
                    super().__init__()
                    self.my_widget = QtWidgets.QWidget()
                    self.my_widget.setWindowTitle("Do you accept?")
                    self.my_widget.setWindowModality(QtCore.Qt.ApplicationModal)
                    self.my_widget.setFixedSize(200, 100)

                    self.my_widget.accept_button = QtWidgets.QPushButton("Accept")
                    self.my_widget.reject_button = QtWidgets.QPushButton("Reject")
                    self.my_widget.layout = QtWidgets.QVBoxLayout()
                    self.my_widget.layout.addWidget(self.my_widget.accept_button)
                    self.my_widget.layout.addWidget(self.my_widget.reject_button)

                    self.app.register_with_qt_future(self.my_function)



                def my_function(self, future: QtFuture, *args, **kwargs):
                    self.my_widget.setText("Do you accept?")
                    self.my_widget.show()

                    self.current_future = future


                def on_accept(self):
                    self.current_future.set_result(True)

                def on_reject(self):
                    self.current_future.set_result(False)


        """
        from rekuest_next.qt.builders import qtwithfutureactifier

        return self.services["rekuest"].register(
            function, actifier=qtwithfutureactifier, **kwargs
        )

    def register_with_qt_generator(self, function: Callable, **kwargs):
        from rekuest_next.qt.builders import qtwithgeneratoractifier

        return self.services["rekuest"].register(
            function, actifier=qtwithgeneratoractifier, **kwargs
        )

    def run(self):
        """Run the app"""
        self.services["rekuest"].run()

    async def __aenter__(self):
        await super().__aenter__()
        for service in self.services.values():
            await service.__aenter__()

        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        for service in self.services.values():
            await service.__aexit__(exc_type, exc_value, traceback)



import contextvars
from rich.console import Console
import rich_click as click
from .types import Manifest


current_console: contextvars.ContextVar[Console] = contextvars.ContextVar(
    "current_console"
)
current_manifest: contextvars.ContextVar[Manifest] = contextvars.ContextVar(
    "current_manifest"
)


def get_console(ctx) -> Console:
    try:
        return ctx.obj["console"]
    except LookupError:
        raise click.ClickException(
            "No Current Console. Probably you are not running this command from the CLI."
        )


def set_console(ctx, console):
    ctx.obj["console"] = console


def get_manifest(ctx) -> Manifest:
    """Get the current manifest or raise an exception."""
    try:
        return ctx.obj["manifest"]
    except LookupError:
        raise click.ClickException(
            "No manifest found. You need to run the `arkitekt_next init` command first."
        )


def set_manifest(ctx, manifest):
    ctx.obj["manifest"] = manifest



from typing import Dict, Any, Protocol
from blok import blok, InitContext, Option
from blok import service
from dataclasses import dataclass


@dataclass
class OllamaCredentials:
    api_key: str
    api_secret: str
    api_url: str


@service("io.ollama.ollama", description=" A self-hosted ollama LLM server")
class OllamaService(Protocol):
    pass



from pydantic import BaseModel
from typing import Dict, Any
from blok import blok, InitContext, Option, ExecutionContext
from blok.tree import YamlFile


class AdminCredentials(BaseModel):
    password: str
    username: str
    email: str


@blok(
    "live.arkitekt.docker_socket", description="Hosting the docker socket of the host"
)
class DockerSocketBlok:
    def __init__(self) -> None:
        self.docker_socket = "/var/run/docker.sock"
        self.registered_configs = {}

    def preflight(self, init: InitContext):
        for key, value in init.kwargs.items():
            setattr(self, key, value)

    def register_socket(self, name: str) -> str:
        self.registered_configs[name] = name
        return self.docker_socket

    def build(self, ex: ExecutionContext):
        pass

    def get_options(self):
        config_path = Option(
            subcommand="docker_socket",
            help="Which docker_socket to use for configs",
            default=self.docker_socket,
            show_default=True,
        )

        return [config_path]



from typing import Dict, Any
import secrets
from arkitekt_next.bloks.services.gateway import GatewayService
from arkitekt_next.bloks.services.mount import MountService
from blok import blok, InitContext, ExecutionContext, Option, Command
from blok.bloks.services.dns import DnsService
from blok.bloks.services.vscode import VsCodeService
from blok.tree import Repo, YamlFile


@blok(
    "live.arkitekt.orkestrator",
    description="Orkestrator in its web-based UI",
)
class OrkestratorBlok:
    def __init__(self) -> None:
        self.dev = False
        self.disable = False
        self.host = "orkestrator"
        self.path_name = "orkestrator"
        self.repo = "https://github.com/arkitektio/orkestrator-next"
        self.build_command = ["yarn"]
        self.up_command = ["yarn", "start"]

    def preflight(
        self,
        init: InitContext,
        gateway: GatewayService,
        mount: MountService,
        vscode: VsCodeService | None = None,
    ):
        for key, value in init.kwargs.items():
            setattr(self, key, value)

        if self.disable:
            return

        if not self.dev:
            gateway.expose(self.path_name, 80, self.host)

            self.service = {
                "image": "jhnnsrs/orkestrator_next:nightly",
                "environment": {
                    "PATH_NAME": self.path_name,
                },
            }

            return

        else:

            self.mount = mount.register_mount("orkestrator", Repo(self.repo))

            if vscode is not None:
                vscode.register_task(
                    "Run Orkestrator",
                    "shell",
                    "yarn",
                    ["start"],
                    {"cwd": f"{self.mount}"},
                )
                vscode.register_task(
                    "Build Orkestrator",
                    "shell",
                    "yarn",
                    [],
                    {"cwd": f"{self.mount}"},
                )

            self.initialized = True

    def build(self, context: ExecutionContext):
        if self.disable:
            return

        if self.dev:

            context.install_commands.set_nested(
                "orkestrator", Command(self.build_command, cwd=self.mount)
            )
            context.up_commands.set_nested(
                "orkestrator", Command(self.up_command, cwd=self.mount)
            )
            return

        context.docker_compose.set_nested("services", self.host, self.service)
        pass

    def get_options(self):
        with_repos = Option(
            subcommand="repo",
            help="The default repo to use for the orkestrator",
            default=self.repo,
        )
        with_disable = Option(
            subcommand="disable",
            help="Should we disable the orkestrator service?",
            default=self.disable,
        )
        with_dev = Option(
            subcommand="dev",
            help="Should we mount orkestrator as dev?",
            default=self.dev,
        )
        with_path_name = Option(
            subcommand="path_name",
            help="The path name for the orkestrator service",
            default=self.path_name,
        )
        with_host = Option(
            subcommand="host",
            help="The host for the orkestrator service",
            default=self.host,
        )

        return [
            with_dev,
            with_repos,
            with_path_name,
            with_disable,
            with_host,
        ]



from typing import Dict, Any, Protocol
from blok import blok, InitContext, Option
from blok import service
from dataclasses import dataclass


@service("live.arkitekt.names")
class NameService(Protocol):

    def retrieve_name(self) -> str:
        pass



from typing import Dict, Any
import secrets

from arkitekt_next.bloks.lok import LokBlok
from arkitekt_next.bloks.services.gateway import GatewayService
from arkitekt_next.bloks.socket import DockerSocketBlok
from blok import blok, InitContext, ExecutionContext, Option
from blok.bloks.services.dns import DnsService
from blok.tree import YamlFile, Repo


@blok(
    "live.arkitekt.internal_engine",
    description="An Arkitekt plugin engine running inside a docker container",
)
class InternalDockerBlok:
    def __init__(self) -> None:
        self.host = "internal_docker"

        self.image = "jhnnsrs/deployer:nightly"
        self.instance_id = "INTERNAL_DOCKER"

    def preflight(
        self,
        init: InitContext,
        gateway: GatewayService,
        lok: LokBlok,
        socket: DockerSocketBlok,
    ):
        for key, value in init.kwargs.items():
            setattr(self, key, value)

        deps = init.dependencies

        if self.skip:
            return

        self._socket = socket.register_socket(self.host)
        self.gateway_host = gateway.get_internal_host()
        self.gateway_port = gateway.get_http_port()

        self.token = lok.retrieve_token()

        self.gateway_network = gateway.retrieve_gateway_network()

        self.command = f"arkitekt-next run prod --redeem-token={self.token} --url http://{self.gateway_host}:{self.gateway_port}"

        self.initialized = True

    def build(self, context: ExecutionContext):
        if self.skip:
            return
        db_service = {
            "image": self.image,
            "command": self.command,
            "volumes": [f"{self._socket}:/var/run/docker.sock"],
            "environment": {
                "INSTANCE_ID": self.instance_id,
                "ARKITEKT_GATEWAY": f"http://{self.gateway_host}:{self.gateway_port}",
                "ARKITEKT_NETWORK": self.gateway_network,
            },
            "deploy": {
                "restart_policy": {
                    "condition": "on-failure",
                    "delay": "10s",
                    "max_attempts": 10,
                    "window": "300s",
                },
            },
        }

        context.docker_compose.set_nested("services", self.host, db_service)

    def get_options(self):
        with_host = Option(
            subcommand="host",
            help="The host of this service",
            default=self.host,
        )
        with_skip = Option(
            subcommand="skip",
            help="Should we skip creating this service?",
            default=False,
            type=bool,
        )

        return [
            with_host,
            with_skip,
        ]

    def __str__(self) -> str:
        return (
            f"InterDocker(host={self.host}, command={self.command}, image={self.image})"
        )



from herre_next.herre import Herre
from fakts_next import Fakts
from herre_next.grants.oauth2.refresh import RefreshGrant
from herre_next.fakts.grant import FaktsGrant
from herre_next.fakts.fakts_qt_store import FaktsQtStore

from herre_next.grants.auto_login import AutoLoginGrant
from herre_next.grants.qt.auto_login import AutoLoginWidget
from herre_next.fakts.fakts_endpoint_fetcher import FaktsUserFetcher
from arkitekt_next.base_models import Manifest, User
from arkitekt_next.apps.service.grant_registry import ARKITEKT_GRANT_REGISTRY


class ArkitektNextAutoLogin(AutoLoginGrant):
    store: FaktsQtStore
    fetcher: FaktsUserFetcher
    grant: FaktsGrant


class ArkitektNextRefreshGrant(RefreshGrant):
    grant: ArkitektNextAutoLogin


class ArkitektNextHerreQt(Herre):
    grant: ArkitektNextRefreshGrant


def build_arkitekt_next_qt_herre_next(
    manifest: Manifest,
    fakts: Fakts,
    login_widget=None,
    parent=None,
    settings=None,
):
    login_widget = login_widget or AutoLoginWidget(parent=parent)

    grant = ArkitektNextAutoLogin(
        store=FaktsQtStore(
            fakts=fakts,
            settings=settings,
            fakts_key="lok.endpoint_url",
        ),
        widget=login_widget,
        fetcher=FaktsUserFetcher(
            fakts=fakts, fakts_key="lok.userinfo_url", userModel=User
        ),
        grant=FaktsGrant(
            fakts=fakts, fakts_group="lok", grant_registry=ARKITEKT_GRANT_REGISTRY
        ),
    )

    return ArkitektNextHerreQt(
        grant=ArkitektNextRefreshGrant(grant=grant),
        fetcher=FaktsUserFetcher(
            fakts=fakts, fakts_key="lok.userinfo_url", userModel=User
        ),
    )



import click
from pydantic import BaseModel
from typing import Dict, Any
import yaml
import secrets
from blok import blok, InitContext

from blok import blok, InitContext, ExecutionContext, Option
from blok.tree import YamlFile, Repo
from arkitekt_next.bloks.base import BaseArkitektService


class AccessCredentials(BaseModel):
    password: str
    username: str
    host: str
    port: str
    db_name: str


@blok("live.arkitekt.elektro", description="Mikro is the bio-image database for arkitekt")
class ElektroBlok(BaseArkitektService):
    def __init__(self) -> None:
        self.dev = False
        self.host = "elektro"
        self.command = "bash run-debug.sh"
        self.repo = "https://github.com/jhnnsrs/elektro-server"
        self.scopes = {
            "mikro_read": "Read image from the database",
            "mikro_write": "Write image to the database",
        }
        self.image = "jhnnsrs/elektro:nightly"
        self.mount_repo = False
        self.build_repo = False
        self.buckets = ["media", "zarr", "parquet"]
        self.secret_key = secrets.token_hex(16)

    def get_builder(self):
        return "arkitekt.generic"

    def build(self, context: ExecutionContext):
        context.docker_compose.set_nested("services", self.host, self.service)



from typing import Any, Dict, Optional

from pydantic import BaseModel

from arkitekt_next.bloks.kraph import KraphBlok
from arkitekt_next.bloks.tailscale import TailscaleBlok
from blok import InitContext, Panel, Renderer, blok

from .fluss import FlussBlok
from .internal_docker import InternalDockerBlok
from .kabinet import KabinetBlok
from .livekit import LocalLiveKitBlok
from .mikro import MikroBlok
from .orkestrator import OrkestratorBlok
from .rekuest import RekuestBlok
from .ollama import OllamaBlok
from .elektro import ElektroBlok


class AdminCredentials(BaseModel):
    password: str
    username: str
    email: str


@blok(
    "live.arkitekt",
    dependencies=[
        MikroBlok.as_dependency(True, True),
        KabinetBlok.as_dependency(True, True),
        RekuestBlok.as_dependency(True, True),
        FlussBlok.as_dependency(True, True),
        InternalDockerBlok.as_dependency(True, True),
        LocalLiveKitBlok.as_dependency(True, True),
        OrkestratorBlok.as_dependency(True, False),
        KraphBlok.as_dependency(True, False),
        TailscaleBlok.as_dependency(True, False),
        OllamaBlok.as_dependency(True, False),
        ElektroBlok.as_dependency(True, False),
    ],
)
class ArkitektBlok:

    def entry(self, renderer: Renderer):
        renderer.render(
            Panel(
                f"""This is the arkitekt build that allows you to setup a full stack arkitekt application. You can use this to setup a full stack application with the following services""",
                expand=False,
                title="Welcome to Arkitekt!",
                style="bold magenta",
            )
        )

    def build(self, cwd):
        pass



def missing_install(name: str, error: Exception):
    def real_missing_install(*args, **kwargs):
        raise ImportError(
            f"Missing import: {name}. Please install the missing package. "
        ) from error

    return real_missing_install


try:
    from rekuest_next.register import register, register_structure
    from rekuest_next.agents.hooks import background
    from rekuest_next.agents.hooks import startup
    from rekuest_next.agents.context import context
    from rekuest_next.state.state import state
    from rekuest_next.actors.reactive.api import abreakpoint, breakpoint
    from rekuest_next.actors.reactive.api import progress, aprogress
    from rekuest_next.actors.reactive.api import log, alog
    from rekuest_next.register import test, benchmark
    from rekuest_next.structures.model import model
    from rekuest_next.utils import call, call_raw, acall, acall_raw, find, afind
    from rekuest_next.define import define
except ImportError as e:
    raise e
    register_structure = missing_install("rekuest_next", e)
    register = missing_install("rekuest_next", e)
    background = missing_install("rekuest_next", e)
    abreakpoint = missing_install("rekuest_next", e)
    breakpoint = missing_install("rekuest_next", e)
    startup = missing_install("rekuest_next", e)
    context = missing_install("rekuest_next", e)
    state = missing_install("rekuest_next", e)
    progress = missing_install("rekuest_next", e)
    aprogress = missing_install("rekuest_next", e)
    log = missing_install("rekuest_next", e)
    alog = missing_install("rekuest_next", e)
    call = missing_install("rekuest_next", e)
    call_raw = missing_install("rekuest_next", e)
    acall = missing_install("rekuest_next", e)
    acall_raw = missing_install("rekuest_next", e)
    find = missing_install("rekuest_next", e)
    afind = missing_install("rekuest_next", e)

from .builders import easy, interactive
from .apps.types import App
from fakts_next.helpers import afakt, fakt
from .init_registry import init, InitHookRegisty, get_current_init_hook_registry
from .service_registry import (
    require,
    ServiceBuilderRegistry,
    get_current_service_registry,
)


__all__ = [
    "App",
    "require",
    "easy",
    "interactive",
    "publicqt",
    "jupy",
    "log",
    "alog",
    "afakt",
    "fakt",
    "progress",
    "aprogress",
    "scheduler",
    "register_structure",
    "requirement",
    "ServiceBuilderRegistry",
    "get_current_service_registry",
    "register",
    "group",
    "useGuardian",
    "useInstanceID",
    "find",
    "afind",
    "call",
    "call_raw",
    "acall",
    "acall_raw",
    "model",
    "test",
    "benchmark",
    "useUser",
    "next",
    "state",
    "context",
    "background",
    "startup",
    "register_next",
    "init",
    "InitHookRegisty",
    "get_current_init_hook_registry",
]



import rich_click as click


import rich_click as click
from .remote import remote


@click.group()
@click.pass_context
def call(ctx):
    """Inspects your arkitekt_next app

    Inspects various parts of your arkitekt_next app. This is useful for debugging
    and development. It also represents methods that are called by the arkitekt_next
    server when you run your app in production mode.

    """


call.add_command(remote, "remote")



from typing import Dict, Any
import secrets

from arkitekt_next.bloks.services.gateway import GatewayService
from arkitekt_next.bloks.services.ollama import OllamaService, OllamaCredentials
from blok import blok, InitContext, ExecutionContext, Option
from blok.tree import YamlFile, Repo


@blok(OllamaService, description="a self-hosted Ollama LLM server")
class OllamaBlok:
    def __init__(self) -> None:
        self.host = "ollama"
        self.image = "ollama/ollama:latest"
        self.port = 11434
        self.skip = False
        self.gpu = bool

    def preflight(self, init: InitContext, gateway: GatewayService):
        for key, value in init.kwargs.items():
            setattr(self, key, value)

        if self.skip:
            return

        gateway.expose_port_to(self.port, self.host, 11434, False)

        self.initialized = True

    def build(self, context: ExecutionContext):
        if self.skip:
            return
        db_service = {
            "labels": [
                "fakts.service=io.ollama.ollama",
                "fakts.builder=ollama.ollama",
            ],
            "image": self.image,
            "environment": {
                "OLLAMA_KEEP_ALIVE": "24h",
            },
        }

        if self.gpu:
            db_service["deploy"] = {
                "resources": {
                    "reservations": {
                        "devices": [
                            {
                                "driver": "nvidia",
                                "count": 1,
                                "capabilities": ["gpu"],
                            }
                        ]
                    }
                }
            }

        context.docker_compose.set_nested("services", self.host, db_service)

    def get_options(self):
        with_host = Option(
            subcommand="host",
            help="The fakts url for connection",
            default=self.host,
        )
        with_skip = Option(
            subcommand="skip",
            help="The fakts_next url for connection",
            default=False,
            type=bool,
        )
        with_gpu = Option(
            subcommand="gpu",
            help="Whether to use a gpu",
            default=False,
            type=self.gpu,
        )

        return [
            with_host,
            with_skip,
            with_gpu,
        ]

    def __str__(self) -> str:
        return (
            f"Ollama(host={self.host}, command={self.command}, image={self.image})"
        )



from enum import Enum
from qtpy import QtWidgets, QtGui, QtCore
from arkitekt_next.qt.types import QtApp
from koil.qt import async_to_qt

from arkitekt_next.apps.types import App
from .utils import get_image_path
from typing import Optional, Callable
import logging
import aiohttp
from logging import LogRecord

logger = logging.getLogger(__name__)


class Logo(QtWidgets.QWidget):
    """Logo widget

    THhe logo widget is a widget that can be used to display a logo in the settings dialog.
    It will download the logo from the url, and display it.


    """

    def __init__(self, url: str, *args, **kwargs) -> None:
        """Logo widget

        Parameters
        ----------
        url : str
            The url to download the logo from.
        """

        super().__init__(*args, **kwargs)
        self.logo_url = url
        self.getter = async_to_qt(
            self.aget_image,
        )  # we use async_to_qt to convert the async function to a Qt signal. (see koil docs)

        self.mylayout = QtWidgets.QVBoxLayout()
        self.setLayout(self.mylayout)
        self.getter.returned.connect(self.on_image)
        self.getter.run()

    def on_image(self, data: Optional[bytes]) -> None:
        """Callback for when the image is downloaded."""
        if not data:
            return

        self.pixmap = QtGui.QPixmap()
        self.pixmap.loadFromData(data)
        self.scaled_pixmap = self.pixmap.scaledToWidth(100)
        self.logo = QtWidgets.QLabel()
        self.logo.setPixmap(self.scaled_pixmap)

        self.mylayout.addWidget(self.logo)

    async def aget_image(self) -> Optional[bytes]:
        """Async function to download the image."""
        async with aiohttp.ClientSession() as session:
            async with session.get(self.logo_url) as resp:
                if resp.status == 200 and "image" in resp.headers["Content-Type"]:
                    data = await resp.read()
                    return data
                else:
                    logger.error(
                        f"Failed to download the image. Status code: {resp.status}"
                    )
                    return None


class ArkitektNextLogsRetriever(logging.Handler, QtCore.QObject):
    """A logging handler that will emit a Qt signal when a log message is received."""

    appendPlainText = QtCore.Signal(str)

    def __init__(self, widget: QtWidgets.QPlainTextEdit, *args, **kwargs) -> None:
        """A logging handler that will emit a Qt signal when a log message is received.

        Parameters
        ----------
        widget : QtWidgets.QPlainTextEdit
            A plain text edit widget to display the logs in.
        """
        super().__init__()
        QtCore.QObject.__init__(self)
        self.appendPlainText.connect(widget.appendPlainText)
        self.setFormatter(
            logging.Formatter(
                "%(asctime)s %(levelname)s %(module)s %(funcName)s %(message)s"
            )
        )

    def emit(self, record: LogRecord) -> None:
        """Emit a Qt signal when a log message is received."""
        msg = self.format(record)
        self.appendPlainText.emit(msg)


class ArkitektNextLogs(QtWidgets.QDialog):
    """A dialog that will display the logs of the app."""

    def __init__(
        self,
        settings: QtCore.QSettings,
        *args,
        log_level_key: str = "log_level",
        log_to_file_key: str = "log_to_file",
        **kwargs,
    ) -> None:
        """A dialog that will display the logs of the app.

        Parameters
        ----------
        settings : QtCore.QSettings
            The settings object to use to store the log level. (so that is persistent,
            and can be changed by the user)
        log_level_key : str, optional
            The key to use to store the log level, by default "log_level"
        log_to_file_key : str, optional
            The key to use to store whether the logs should be written to a file,
             by default "log_to_file"
        """
        super().__init__(*args, **kwargs)
        self.log_level_key = log_level_key
        self.log_to_file_key = log_to_file_key
        self.settings = settings
        self.setWindowTitle("Logs")
        self.mylayout = QtWidgets.QVBoxLayout()
        self.text = QtWidgets.QPlainTextEdit(parent=self)
        self.text.setMaximumBlockCount(5000)
        self.text.setReadOnly(True)
        self.mylayout.addWidget(self.text)
        self.logRetriever = ArkitektNextLogsRetriever(self.text)
        logging.getLogger().addHandler(self.logRetriever)
        logging.getLogger().setLevel(self.log_level)
        self.setLayout(self.mylayout)

    def update_log_level(self, level: str) -> None:
        """Update the log level.

        Parameters
        ----------
        level : str
           Update the log level to this level.
        """
        logging.getLogger().setLevel(level)
        self.settings.setValue(self.log_level_key, level)

    @property
    def log_to_file(self) -> bool:
        """Should the logs be written to a file."""
        return self.settings.value(self.log_to_file_key, False, bool)

    @property
    def log_level(self) -> str:
        """The log level in use."""
        return self.settings.value(self.log_level_key, "INFO", str)


class Profile(QtWidgets.QDialog):
    """The profile dialog.

    It will display the logo, and allow the user to change the user, and server.
    It will also allow the user to change the log level, and show the logs on
    demand.


    """

    updated = QtCore.Signal()

    def __init__(
        self,
        app: App,
        bar: "MagicBar",
        *args,
        dark_mode: bool = False,
        **kwargs,
    ) -> None:
        """The profile dialog.

        Parameters
        ----------
        app : QtApp
            The app to use. (needs to be a QtApp as it uses Qt signals)
        bar : MagicBar
            The magic bar to use and to update when the user changes the settings.
        dark_mode : bool, optional
            Should we use dark_mode, by default False
            TODO: implement dark mode
        """
        super().__init__(*args, **{"parent": bar, **kwargs})
        self.app = app
        self.bar = bar

        self.settings = QtCore.QSettings(
            "arkitekt_next",
            f"{self.app.manifest.identifier}:{self.app.manifest.version}:profile",
        )

        self.setWindowTitle("Settings")

        self.infobar = QtWidgets.QVBoxLayout()

        self.mylayout = QtWidgets.QHBoxLayout()
        self.setLayout(self.mylayout)
        self.mylayout.addLayout(self.infobar)

        if self.app.manifest.logo:
            self.infobar.addWidget(Logo(self.app.manifest.logo, parent=self))

        self.infobar.addWidget(QtWidgets.QLabel(self.app.manifest.identifier))
        self.infobar.addWidget(QtWidgets.QLabel(self.app.manifest.version))

        self.unkonfigure_button = QtWidgets.QPushButton("Reconnect")
        self.unkonfigure_button.clicked.connect(lambda: self.bar.refresh_task.run())

        button_bar = QtWidgets.QHBoxLayout()
        self.infobar.addLayout(button_bar)
        button_bar.addWidget(self.unkonfigure_button)

        self.logs = ArkitektNextLogs(self.settings, parent=self)

        self.go_all_the_way_button = QtWidgets.QPushButton("One click provide")
        self.go_all_the_way_button.setCheckable(True)
        self.go_all_the_way_button.setChecked(self.go_all_the_way_down)
        self.go_all_the_way_button.clicked.connect(self.on_go_all_the_way_clicked)

        self.sidebar = QtWidgets.QVBoxLayout()
        self.mylayout.addLayout(self.sidebar)

        self.show_logs_button = QtWidgets.QPushButton("Show Logs")
        self.show_logs_button.clicked.connect(self.logs.show)

        self.sidebar.addWidget(self.go_all_the_way_button)
        self.sidebar.addWidget(self.show_logs_button)
        self.sidebar.addStretch()

    def on_go_all_the_way_clicked(self, checked: bool) -> None:
        """Callback for when the go all the way button is clicked.

        Will update the settings, and emit the updated signal.

        Parameters
        ----------
        checked : bool
            Whether the button is checked or not.

        """
        self.settings.setValue("go_all_the_way_down", checked)
        self.updated.emit()

    @property
    def go_all_the_way_down(self) -> bool:
        """Should the app go all the way down when the user clicks the magic button."""
        return self.settings.value("go_all_the_way_down", True, bool)


class AppState(str, Enum):
    """The state of the app."""

    READY = "ready"
    DOWN = "down"
    UP = "up"


class ProcessState(str, Enum):
    UNKONFIGURED = "unkonfigured"
    UNLOGGED = "unlogged"
    UNPROVIDED = "unprovided"
    PROVIDING = "providing"


class MagicBar(QtWidgets.QWidget):
    """Magic bar widget.

    The magic bar is a small button widget, that can be used to configure, login and put the
    app in providing and non providing states.. It also has a gear button that opens the
    profile dialog. To adjust some parameters of the app"""

    CONNECT_LABEL = "Connect"

    app_state_changed = QtCore.Signal()
    app_up = QtCore.Signal()
    app_down = QtCore.Signal()
    app_error = QtCore.Signal()
    state = AppState.DOWN
    process_state = ProcessState.UNKONFIGURED

    def __init__(
        self,
        app: QtApp,
        dark_mode: bool = False,
        on_error: Optional[Callable[[Exception], None]] = None,
    ) -> None:
        """Magic bar Widget

        This widget is a small button widget, that can be used to configure, login and put the
        app in providing and non providing states.. It also has a gear button that opens the
        profile dialog, that can be used to adjust some parameters of the qt app.

        Parameters
        ----------
        app : QtApp
            A qt app to use.
        dark_mode : bool, optional
            Should we use the dark mode, by default False
        on_error : Optional[Callable[[Exception], None]], optional
            And additinal callback if an error is raised, by default None
        """
        super().__init__()
        self.app = app

        # assert isinstance(
        #     self.app.koil, QtPedanticKoil
        # ), f"Koil should be Qt Koil but is {type(self.app.koil)}"
        self.dark_mode = dark_mode

        self.profile = Profile(app, self, dark_mode=dark_mode)
        self.profile.updated.connect(self.on_profile_updated)

        self.configure_task = async_to_qt(self.app.fakts.aget)
        self.configure_task.errored.connect(self.configure_errored)
        self.configure_task.returned.connect(self.set_unlogined)

        self.refresh_task = async_to_qt(self.app.fakts.arefresh)
        self.refresh_task.errored.connect(self.configure_errored)
        self.refresh_task.returned.connect(self.set_unlogined)

        self.get_token_task = async_to_qt(self.app.herre.aget_token)
        self.get_token_task.errored.connect(self.login_errored)
        self.get_token_task.returned.connect(self.set_unprovided)

        self.refresh_token_task = async_to_qt(self.app.herre.arefresh_token)
        self.refresh_token_task.started.connect(self.set_providing)
        self.refresh_token_task.errored.connect(self.login_errored)
        self.refresh_token_task.returned.connect(self.set_unprovided)

        self.provide_task = async_to_qt(self.app.services.get("rekuest").agent.aprovide)
        self.provide_task.errored.connect(self.provide_errored)
        self.provide_task.returned.connect(self.set_unprovided)

        self.magicb = QtWidgets.QPushButton(MagicBar.CONNECT_LABEL)
        self.magicb.setMinimumHeight(30)
        self.magicb.setMaximumHeight(30)

        self.configure_future = None
        self.login_future = None
        self.provide_future = None

        self.mylayout = QtWidgets.QHBoxLayout()
        self.gearb_pix = QtGui.QPixmap(get_image_path("gear.png", dark_mode=dark_mode))
        self.gearb = QtWidgets.QPushButton()
        self.gearb.setIcon(QtGui.QIcon(self.gearb_pix))
        self.gearb.setMinimumWidth(30)
        self.gearb.setMaximumWidth(30)
        self.gearb.setMinimumHeight(30)
        self.gearb.setMaximumHeight(30)
        self._on_error = on_error

        self.magicb.clicked.connect(self.magic_button_clicked)
        self.gearb.clicked.connect(self.gear_button_clicked)

        self.mylayout.addWidget(self.magicb)
        self.mylayout.addWidget(self.gearb)
        self.setLayout(self.mylayout)

        self.set_unkonfigured()
        self.on_profile_updated()

    def on_profile_updated(self) -> None:
        """Callback for when the profile is updated."""
        if self.profile.go_all_the_way_down:
            self.set_unprovided()

    def show_error(self, ex: Exception) -> None:
        """Show an error message

        Parameters
        ----------
        ex : Exception
            The exception to show.
        """
        if self._on_error:
            self._on_error(ex)
        else:
            logger.error(f"Error {repr(ex)}")

    def task_errored(self, ex: Exception) -> None:
        """_summary_

        Parameters
        ----------
        ex : Exception
            _description_

        Raises
        ------
        ex
            _description_
        """
        raise ex

    def configure_errored(self, ex: Exception) -> None:
        self.set_unkonfigured()
        self.show_error(ex)

    def login_errored(self, ex: Exception) -> None:
        self.set_unlogined()
        self.show_error(ex)

    def provide_errored(self, ex: Exception) -> None:
        self.set_unprovided()
        self.show_error(ex)

    def on_configured(self) -> None:
        self.magicb.setText("Login")

    def on_login(self) -> None:
        self.magicb.setText("Provide")

    def on_provided(self) -> None:
        self.magicb.setText("Provide ended")

    def on_providing_ended(self) -> None:
        pass

    def gear_button_clicked(self) -> None:
        self.profile.show()

    def update_movie(self) -> None:
        self.magicb.setIcon(QtGui.QIcon(self.magicb_movie.currentPixmap()))

    def set_button_movie(self, movie) -> None:
        self.magicb_movie = QtGui.QMovie(
            get_image_path(movie, dark_mode=self.dark_mode)
        )
        self.magicb_movie.frameChanged.connect(self.update_movie)
        self.magicb_movie.setScaledSize(QtCore.QSize(30, 30))
        self.magicb_movie.start()

    def set_unkonfigured(self) -> None:
        self.state = AppState.DOWN
        self.process_state = ProcessState.UNKONFIGURED
        self.app_down.emit()
        self.app_state_changed.emit()
        self.set_button_movie("pink pulse.gif")
        self.profile.unkonfigure_button.setDisabled(True)
        self.magicb.setDisabled(False)
        self.magicb.setText("Konfigure App")

    def set_unlogined(self) -> None:
        self.state = AppState.DOWN
        self.process_state = ProcessState.UNLOGGED
        self.app_down.emit()

        self.app_state_changed.emit()
        self.set_button_movie("orange pulse.gif")
        self.profile.unkonfigure_button.setDisabled(False)
        self.magicb.setDisabled(False)
        self.magicb.setText("Login")

    def set_unprovided(self) -> None:
        self.state = AppState.UP
        self.process_state = ProcessState.UNPROVIDED
        self.app_up.emit()

        self.app_state_changed.emit()
        self.set_button_movie("green pulse.gif")
        self.profile.unkonfigure_button.setDisabled(False)
        self.magicb.setDisabled(False)
        self.magicb.setText("Provide")

    def set_providing(self) -> None:
        self.state = AppState.UP
        self.process_state = ProcessState.PROVIDING
        self.app_up.emit()

        self.app_state_changed.emit()
        self.set_button_movie("red pulse.gif")
        self.profile.unkonfigure_button.setDisabled(False)
        self.magicb.setDisabled(False)
        self.magicb.setText("Cancel Provide..")

    def magic_button_clicked(self) -> None:
        if (
            self.process_state == ProcessState.UNKONFIGURED
            and not self.profile.go_all_the_way_down
        ):
            if not self.configure_future or self.configure_future.done():
                self.configure_future = self.configure_task.run()
                self.magicb.setText("Cancel Configuration")
                return
            if not self.configure_future.done():
                self.configure_future.cancel()
                self.set_unkonfigured()
                return

        if (
            self.process_state == ProcessState.UNLOGGED
            and not self.profile.go_all_the_way_down
        ):
            if not self.login_future or self.login_future.done():
                self.login_future = self.get_token_task.run()
                self.magicb.setText("Cancel Login")
                return
            if not self.login_future.done():
                self.login_future.cancel()
                self.magicb.setText("Login")
                self.set_unlogined()
                return

        if (
            self.process_state != ProcessState.PROVIDING
            or self.profile.go_all_the_way_down
        ):
            if not self.provide_future or self.provide_future.done():
                self.provide_future = self.provide_task.run()
                self.set_providing()
                return

        if self.provide_future:
            if not self.provide_future.done():
                self.provide_future.cancel()
                self.set_unprovided()
                return



from arkitekt_next.apps.service.fakts_next import (
    build_arkitekt_next_redeem_fakts_next,
)
from arkitekt_next.apps.service.fakts_qt import build_arkitekt_next_qt_fakts_next
from arkitekt_next.apps.service.herre_qt import build_arkitekt_next_qt_herre_next
from arkitekt_next.utils import create_arkitekt_next_folder
from arkitekt_next.base_models import Manifest
from arkitekt_next.apps.types import App
from arkitekt_next.service_registry import (
    ServiceBuilderRegistry,
    get_default_service_registry,
)
from arkitekt_next.constants import DEFAULT_ARKITEKT_URL
from qtpy import QtWidgets, QtCore
from typing import List, Optional
import os
import logging
from typing import List, Optional
import logging
import os
from arkitekt_next.qt.types import QtApp
from arkitekt_next.apps.service.fakts_next import (
    build_arkitekt_next_fakts_next,
    build_arkitekt_next_redeem_fakts_next,
    build_arkitekt_next_token_fakts_next,
)
from arkitekt_next.apps.service.herre import build_arkitekt_next_herre_next
from arkitekt_next.utils import create_arkitekt_next_folder
from arkitekt_next.base_models import Manifest
from arkitekt_next.apps.types import App
from arkitekt_next.service_registry import (
    ServiceBuilderRegistry,
)
from arkitekt_next.constants import DEFAULT_ARKITEKT_URL


def devqt(
    identifier: str,
    version: str = "0.0.1",
    logo: Optional[str] = None,
    scopes: Optional[List[str]] = None,
    url: str = DEFAULT_ARKITEKT_URL,
    headless: bool = False,
    log_level: str = "ERROR",
    parent: Optional[QtWidgets.QWidget] = None,
    token: Optional[str] = None,
    no_cache: bool = False,
    redeem_token: Optional[str] = None,
    app_kind: str = "development",
    registry: Optional[ServiceBuilderRegistry] = None,
    description: Optional[str] = None,
    **kwargs,
) -> QtApp:
    """Creates a next app

    A simple way to create an ArkitektNext Next app, ArkitektNext next apps are
    development apps by default, as they will try to register themselves
    with services that are not yet available in production (such as the
    rekuest_next and mikro_next services). They represent the next generation
    of ArkitektNext apps, and will be the default way to create ArkitektNext apps
    in the future. From here be dragons.

    A few things to note:
        -   The Next builder closely mimics the easy builder, but will use the
            next generation of services (such as rekuest_next and mikro_next)
            and will therefore not be compatible with the current generation.

        -  Next apps will try to establish themselves a "development" apps, by default
            which means that they will be authenticated with the ArkitektNext server on
            a per user basis. If you want to create a "desktop" app, which multiple users
            can use, you should set the `app_kind` to "desktop" TODO: Currently not implemented (use next app for this)
        -  The Next builder can also be used in plugin apps, and when provided with a fakts_next token
           will be able to connect to the ArkitektNext server without any user interaction.


    Parameters
    ----------
    identifier : str
        The apps identifier (should be globally unique, see Manifest for more info)
    version : str, optional
        The version of the app, by default "0.0.1"
    logo : str, optional
        The logo of the app as a public http url, by default None
    scopes : List[str], optional
        The scopes, that this apps requires, will default to standard scopes, by default None
    url : str, optional
        The fakts_next server that will be used to configure this app, in a default ArkitektNext deployment this
        is the address of the "Lok Service" (which provides the Fakts API), by default DEFAULT_ARKITEKT_URL
        Will be overwritten by the FAKTS_URL environment variable
    headless : bool, optional
        Should we run in headless, mode, e.g printing necessary interaction into the console (will forexample
        stop opening browser windows), by default False
    log_level : str, optional
        The log-level to use, by default "ERROR"
    token : str, optional
        A fakts_next token to use, by default None
        Will be overwritten by the FAKTS_TOKEN environment variable
    no_cache : bool, optional
        Should we skip caching token, acess-token, by default False
        Attention: If this is set to True, the app will always have to be configured
        and authenticated.
    instance_id : str, optional
        The instance_id to use, by default "main"
        Can be set to a different value, if you want to run multiple intstances
        of the same app by the same user.
        Will be overwritten by the REKUEST_INSTANCE_ID environment variable
    register_reaktion : bool, optional
        Should we register the reaktion extension, by default True
        If set to False, the app will not be able to use the reaktion extension
        (which is necessary for scheduling in app` workflows from fluss)
    app_kind : str, optional
        The kind of app to create, by default "development"
        Can be set to "desktop" to create a desktop app, that can be used by multiple users.

    Returns
    -------
    NextApp
        A built app, that can be used to interact with the ArkitektNext server
    """
    registry = registry or get_default_service_registry()

    url = os.getenv("FAKTS_URL", url)
    token = os.getenv("FAKTS_TOKEN", token)

    manifest = Manifest(
        version=version,
        identifier=identifier,
        scopes=scopes if scopes else ["openid"],
        logo=logo,
        requirements=registry.get_requirements(),
        description=description,
    )
    if token:
        fakts_next = build_arkitekt_next_token_fakts_next(
            manifest=manifest,
            token=token,
            url=url,
        )

    elif redeem_token:
        fakts_next = build_arkitekt_next_redeem_fakts_next(
            manifest=manifest,
            redeem_token=redeem_token,
            url=url,
            no_cache=no_cache,
            headless=headless,
        )
    else:
        fakts_next = build_arkitekt_next_fakts_next(
            manifest=manifest,
            url=url,
            no_cache=no_cache,
            headless=headless,
            client_kind=app_kind,
        )

    herre_next = build_arkitekt_next_herre_next(fakts_next=fakts_next)

    params = kwargs

    create_arkitekt_next_folder(with_cache=True)

    try:
        from rich.logging import RichHandler

        logging.basicConfig(level=log_level, handlers=[RichHandler()])
    except ImportError:
        logging.basicConfig(level=log_level)

    app = QtApp(
        fakts=fakts_next,
        herre=herre_next,
        manifest=manifest,
        services=registry.build_service_map(
            fakts=fakts_next, herre=herre_next, params=params, manifest=manifest
        ),
    )

    app.enter()

    return app


def publicqt(
    identifier: str,
    version: str = "latest",
    logo: Optional[str] = None,
    scopes: Optional[List[str]] = None,
    log_level: str = "ERROR",
    registry: Optional[ServiceBuilderRegistry] = None,
    parent: Optional[QtWidgets.QWidget] = None,
    beacon_widget: Optional[QtWidgets.QWidget] = None,
    login_widget: Optional[QtWidgets.QWidget] = None,
    settings: Optional[QtCore.QSettings] = None,
    **kwargs,
) -> QtApp:
    """Public QtApp creation

    A simple way to create an Arkitekt app with a public grant (allowing users to sign
    in with the application ) utlizing a retrieve grant (necessating a previous configuration
    of the application on the server side)

    Args:
        identifier (str): The apps identifier
        version (str, optional): The apps verion. Defaults to "latest".
        parent (QtWidget, optional): The QtParent (for the login and server select widget). Defaults to None.

    Returns:
        Arkitekt: The Arkitekt app
    """

    registry = registry or check_and_import_services()

    settings = settings or QtCore.QSettings("arkitekt_next", f"{identifier}:{version}")

    manifest = Manifest(
        version=version,
        identifier=identifier,
        scopes=scopes if scopes else ["openid"],
        logo=logo,
        requirements=registry.get_requirements(),
    )

    fakts_next = build_arkitekt_next_qt_fakts_next(
        manifest=manifest,
        beacon_widget=beacon_widget,
        parent=parent,
        settings=settings,
    )

    herre_next = build_arkitekt_next_qt_herre_next(
        manifest,
        fakts=fakts_next,
        login_widget=login_widget,
        parent=parent,
        settings=settings,
    )

    params = kwargs

    try:
        from rich.logging import RichHandler

        logging.basicConfig(level=log_level, handlers=[RichHandler()])
    except ImportError:
        logging.basicConfig(level=log_level)

    app = QtApp(
        fakts=fakts_next,
        herre=herre_next,
        manifest=manifest,
        services=registry.build_service_map(
            fakts=fakts_next, herre=herre_next, params=params, manifest=manifest
        ),
    )

    app.enter()

    return app
